# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)

--

% Perfect reconstruction sliCQ

% right now, even slice length (sl_len) is required. Parameters are the
% same as NSGTF plus slice length, minimal required window length, 
% Q-factor variation, and test run parameters.
"""

import numpy as np
from itertools import cycle, chain, tee
from math import ceil

from .slicing import slicing
from .unslicing import unslicing
from .nsdual import nsdual
from .nsgfwin_sl import nsgfwin
from .nsgtf import nsgtf_sl
from .nsigtf import nsigtf_sl
from .util import calcwinrange
from .fscale import OctScale

# one of the more expensive functions (32/400)
def arrange(cseq, M, fwd):
    cseq = iter(cseq)
    try:
        c0 = next(cseq)  # grab first stream element
    except StopIteration:
        return iter(())
    cseq = chain((c0,), cseq)  # push it back in
    M = list(map(len, c0[0]))  # read off M from the coefficients
    ixs = (
           [(slice(3*mkk//4, mkk), slice(0, 3*mkk//4)) for mkk in M],  # odd
           [(slice(mkk//4, mkk), slice(0, mkk//4)) for mkk in M]  # even
    )
    if fwd:
        ixs = cycle(ixs)
    else:
        ixs = cycle(ixs[::-1])

    return ([
                [np.concatenate((ckk[ix0],ckk[ix1]))
                   for ckk,(ix0,ix1) in zip(ci, ixi)
                ]
             for ci in cci
             ]
             for cci,ixi in zip(cseq, ixs)
            )


def starzip(iterables):
    def inner(itr, i):
        for t in itr:
            yield t[i]
    iterables = iter(iterables)
    it = next(iterables)  # we need that to determine the length of one element
    iterables = chain((it,), iterables)
    return [inner(itr, i) for i,itr in enumerate(tee(iterables, len(it)))]


def chnmap(gen, seq):
    chns = starzip(seq) # returns a list of generators (one for each channel)
    gens = list(map(gen, chns)) # generators including transformation
    return zip(*gens)  # packing channels to one generator yielding channel tuples


class NSGT_sliced:
    def __init__(self, scale, sl_len, tr_area, fs,
                 min_win=16, Qvar=1,
                 real=False, recwnd=False, matrixform=False, reducedform=0,
                 multichannel=False,
                 measurefft=False,
                 multithreading=False,
                 dtype=float):
        assert fs > 0
        assert sl_len > 0
        assert tr_area >= 0
        assert sl_len > tr_area*2
        assert min_win > 0
        assert 0 <= reducedform <= 2

        assert sl_len%4 == 0
        assert tr_area%2 == 0

        self.sl_len = sl_len
        self.tr_area = tr_area
        self.fs = fs
        self.real = real
        self.measurefft = measurefft
        self.multithreading = multithreading
        self.userecwnd = recwnd
        self.reducedform = reducedform

        self.scale = scale
        self.frqs,self.q = self.scale()

        self.g,self.rfbas,self.M = nsgfwin(self.frqs, self.q, self.fs, self.sl_len, sliced=True, min_win=min_win, Qvar=Qvar, dtype=dtype)
        
#        print "rfbas",self.rfbas/float(self.sl_len)*self.fs
        if real:
            assert 0 <= reducedform <= 2
            sl = slice(reducedform,len(self.g)//2+1-reducedform)
        else:
            sl = slice(0,None)

        # coefficients per slice
        self.ncoefs = max(int(ceil(float(len(gii))/mii))*mii for mii,gii in zip(self.M[sl],self.g[sl]))
        
        if matrixform:
            if self.reducedform:
                rm = self.M[self.reducedform:len(self.M)//2+1-self.reducedform]
                self.M[:] = rm.max()
            else:
                self.M[:] = self.M.max()
                
        if multichannel:
            self.channelize = lambda seq: seq
            self.unchannelize = lambda seq: seq
        else:
            self.channelize = lambda seq: ((it,) for it in seq)
            self.unchannelize = lambda seq: (it[0] for it in seq)

        self.wins,self.nn = calcwinrange(self.g, self.rfbas, self.sl_len)
        
        self.gd = nsdual(self.g, self.wins, self.nn, self.M)
        
        self.fwd = lambda fc: nsgtf_sl(fc, self.g, self.wins, self.nn, self.M, real=self.real, reducedform=self.reducedform, measurefft=self.measurefft, multithreading=self.multithreading)
        self.bwd = lambda cc: nsigtf_sl(cc, self.gd, self.wins, self.nn, self.sl_len ,real=self.real, reducedform=self.reducedform, measurefft=self.measurefft, multithreading=self.multithreading)

    @property
    def coef_factor(self):
        return float(self.ncoefs)/self.sl_len
    
    @property
    def slice_coefs(self):
        return self.ncoefs
    
    def forward(self, sig):
        'transform - s: iterable sequence of sequences' 

        sig = self.channelize(sig)

        # Compute the slices (zero-padded Tukey window version)
        f_sliced = slicing(sig, self.sl_len, self.tr_area)
        
        cseq = chnmap(self.fwd, f_sliced)
    
        cseq = arrange(cseq, self.M, True)
        
        cseq = self.unchannelize(cseq)
        
        return cseq


    def backward(self, cseq):
        'inverse transform - c: iterable sequence of coefficients'
                
        cseq = self.channelize(cseq)
        
        cseq = arrange(cseq, self.M, False)

        frec_sliced = chnmap(self.bwd, cseq)
        
        # Glue the parts back together
        ftype = float if self.real else complex
        sig = unslicing(frec_sliced, self.sl_len, self.tr_area, dtype=ftype, usewindow=self.userecwnd)
        
        sig = self.unchannelize(sig)
        
        # discard first two blocks (padding)
        next(sig)
        next(sig)
        return sig


class CQ_NSGT_sliced(NSGT_sliced):
    def __init__(self, fmin, fmax, bins, sl_len, tr_area, fs, min_win=16, Qvar=1, real=False, recwnd=False, matrixform=False, reducedform=0, multichannel=False, measurefft=False, multithreading=False):
        assert fmin > 0
        assert fmax > fmin
        assert bins > 0

        self.fmin = fmin
        self.fmax = fmax
        self.bins = bins  # bins per octave

        scale = OctScale(fmin, fmax, bins)
        NSGT_sliced.__init__(self, scale, sl_len, tr_area, fs, min_win, Qvar, real, recwnd, matrixform=matrixform, reducedform=reducedform, multichannel=multichannel, measurefft=measurefft, multithreading=multithreading)



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)

---

Unit test module
"""



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)
"""



def nsigtf_loop(loopparams, fr, fc):
    fr[:] = 0.
    # The overlap-add procedure including multiplication with the synthesis windows
    # TODO: stuff loop into theano
    for t,(gdii,wr1,wr2,sl1,sl2,temp) in zip(fc, loopparams):
        t1 = temp[sl1]
        t2 = temp[sl2]
        t1[:] = t[sl1]
        t2[:] = t[sl2]
        temp *= gdii
        temp *= len(t)

        fr[wr1] += t2
        fr[wr2] += t1

    return fr



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)
"""

import numpy as np

class Scale:
    dbnd = 1.e-8
    
    def __init__(self, bnds):
        self.bnds = bnds
        
    def __len__(self):
        return self.bnds
    
    def Q(self, bnd=None):
        # numerical differentiation (if self.Q not defined by sub-class)
        if bnd is None:
            bnd = np.arange(self.bnds)
        return self.F(bnd)*self.dbnd/(self.F(bnd+self.dbnd)-self.F(bnd-self.dbnd))
    
    def __call__(self):
        f = np.array([self.F(b) for b in range(self.bnds)],dtype=float)
        q = np.array([self.Q(b) for b in range(self.bnds)],dtype=float)
        return f,q


class OctScale(Scale):
    def __init__(self, fmin, fmax, bpo, beyond=0):
        """
        @param fmin: minimum frequency (Hz)
        @param fmax: maximum frequency (Hz)
        @param bpo: bands per octave (int)
        @param beyond: number of frequency bands below fmin and above fmax (int)
        """
        lfmin = np.log2(fmin)
        lfmax = np.log2(fmax)
        bnds = int(np.ceil((lfmax-lfmin)*bpo))+1
        Scale.__init__(self, bnds+beyond*2)
        odiv = (lfmax-lfmin)/(bnds-1)
        lfmin_ = lfmin-odiv*beyond
        lfmax_ = lfmax+odiv*beyond
        self.fmin = 2**lfmin_
        self.fmax = 2**lfmax_
        self.pow2n = 2**odiv
        self.q = np.sqrt(self.pow2n)/(self.pow2n-1.)/2.
        
    def F(self, bnd=None):
        return self.fmin*self.pow2n**(bnd if bnd is not None else np.arange(self.bnds))
    
    def Q(self, bnd=None):
        return self.q


class LogScale(Scale):
    def __init__(self, fmin, fmax, bnds, beyond=0):
        """
        @param fmin: minimum frequency (Hz)
        @param fmax: maximum frequency (Hz)
        @param bnds: number of frequency bands (int)
        @param beyond: number of frequency bands below fmin and above fmax (int)
        """
        Scale.__init__(self, bnds+beyond*2)
        lfmin = np.log2(fmin)
        lfmax = np.log2(fmax)
        odiv = (lfmax-lfmin)/(bnds-1)
        lfmin_ = lfmin-odiv*beyond
        lfmax_ = lfmax+odiv*beyond
        self.fmin = 2**lfmin_
        self.fmax = 2**lfmax_
        self.pow2n = 2**odiv
        self.q = np.sqrt(self.pow2n)/(self.pow2n-1.)/2.
        
    def F(self, bnd=None):
        return self.fmin*self.pow2n**(bnd if bnd is not None else np.arange(self.bnds))
    
    def Q(self, bnd=None):
        return self.q
    

class LinScale(Scale):
    def __init__(self, fmin, fmax, bnds, beyond=0):
        """
        @param fmin: minimum frequency (Hz)
        @param fmax: maximum frequency (Hz)
        @param bnds: number of frequency bands (int)
        @param beyond: number of frequency bands below fmin and above fmax (int)
        """
        self.df = float(fmax-fmin)/(bnds-1)
        Scale.__init__(self, bnds+beyond*2)
        self.fmin = float(fmin)-self.df*beyond
        if self.fmin <= 0:
            raise ValueError("Frequencies must be > 0.")
        self.fmax = float(fmax)+self.df*beyond

    def F(self, bnd=None):
        return (bnd if bnd is not None else np.arange(self.bnds))*self.df+self.fmin

    def Q(self, bnd=None):
        return self.F(bnd)/(self.df*2)


def hz2mel(f):
    "\cite{shannon:2003}"
    return np.log10(f/700.+1.)*2595.


def mel2hz(m):
    "\cite{shannon:2003}"
    return (np.power(10.,m/2595.)-1.)*700.


class MelScale(Scale):
    def __init__(self, fmin, fmax, bnds, beyond=0):
        """
        @param fmin: minimum frequency (Hz)
        @param fmax: maximum frequency (Hz)
        @param bnds: number of frequency bands (int)
        @param beyond: number of frequency bands below fmin and above fmax (int)
        """
        mmin = hz2mel(fmin)
        mmax = hz2mel(fmax)
        Scale.__init__(self, bnds+beyond*2)
        self.fmin = float(fmin)
        self.fmax = float(fmax)
        self.mbnd = (mmax-mmin)/(bnds-1)  # mels per band
        self.mmin = mmin-self.mbnd*beyond
        self.mmax = mmax+self.mbnd*beyond
        
    def F(self, bnd=None):
        if bnd is None:
            bnd = np.arange(self.bnds)
        return mel2hz(bnd*self.mbnd+self.mmin)

    def Q1(self, bnd=None): # obviously not exact
        if bnd is None:
            bnd = np.arange(self.bnds)
        mel = bnd*self.mbnd+self.mmin
        odivs = (np.exp(mel/-1127.)-1.)*(-781.177/self.mbnd)
        pow2n = np.power(2, 1./odivs)
        return np.sqrt(pow2n)/(pow2n-1.)/2.
    


# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)
"""

import numpy as np
from math import exp, floor, ceil, pi


def hannwin(l):
    r = np.arange(l,dtype=float)
    r *= np.pi*2./l
    r = np.cos(r)
    r += 1.
    r *= 0.5
    return r

def blackharr(n,l=None,mod=True):
    if l is None: 
        l = n
    nn = (n//2)*2
    k = np.arange(n)
    if not mod:
        bh = 0.35875 - 0.48829*np.cos(k*(2*pi/nn)) + 0.14128*np.cos(k*(4*pi/nn)) -0.01168*np.cos(k*(6*pi/nn))
    else:
        bh = 0.35872 - 0.48832*np.cos(k*(2*pi/nn)) + 0.14128*np.cos(k*(4*pi/nn)) -0.01168*np.cos(k*(6*pi/nn))
    bh = np.hstack((bh,np.zeros(l-n,dtype=bh.dtype)))
    bh = np.hstack((bh[-n//2:],bh[:-n//2]))
    return bh

def blackharrcw(bandwidth,corr_shift):
    flip = -1 if corr_shift < 0 else 1
    corr_shift *= flip
    
    M = np.ceil(bandwidth/2+corr_shift-1)*2
    win = np.concatenate((np.arange(M//2,M), np.arange(0,M//2)))-corr_shift
    win = (0.35872 - 0.48832*np.cos(win*(2*np.pi/bandwidth))+ 0.14128*np.cos(win*(4*np.pi/bandwidth)) -0.01168*np.cos(win*(6*np.pi/bandwidth)))*(win <= bandwidth)*(win >= 0)

    return win[::flip],M


def cont_tukey_win(n, sl_len, tr_area):
    g = np.arange(n)*(sl_len/float(n))
    g[np.logical_or(g < sl_len/4.-tr_area/2., g > 3*sl_len/4.+tr_area/2.)] = 0.
    g[np.logical_and(g > sl_len/4.+tr_area/2., g < 3*sl_len/4.-tr_area/2.)] = 1.
    #
    idxs = np.logical_and(g >= sl_len/4.-tr_area/2., g <= sl_len/4.+tr_area/2.)
    temp = g[idxs]
    temp -= sl_len/4.+tr_area/2.
    temp *= pi/tr_area
    g[idxs] = np.cos(temp)*0.5+0.5
    #
    idxs = np.logical_and(g >= 3*sl_len/4.-tr_area/2., g <= 3*sl_len/4.+tr_area/2.)
    temp = g[idxs]
    temp += -3*sl_len/4.+tr_area/2.
    temp *= pi/tr_area
    g[idxs] = np.cos(temp)*0.5+0.5
    #
    return g

def tgauss(ess_ln, ln=0):
    if ln < ess_ln: 
        ln = ess_ln
    #
    g = np.zeros(ln, dtype=float)
    sl1 = int(floor(ess_ln/2))
    sl2 = int(ceil(ess_ln/2))+1
    r = np.arange(-sl1, sl2) # (-floor(ess_len/2):ceil(ess_len/2)-1)
    r = np.exp((r*(3.8/ess_ln))**2*-pi)
    r -= exp(-pi*1.9**2)
    #
    g[-sl1:] = r[:sl1]
    g[:sl2] = r[-sl2:]
    return g

def _isseq(x):
    try:
        len(x)
    except TypeError:
        return False
    return True        

def chkM(M, g):
    if M is None:
        M = np.array(list(map(len, g)))
    elif not _isseq(M):
        M = np.ones(len(g), dtype=int)*M
    return M

def calcwinrange(g, rfbas, Ls):
    shift = np.concatenate(((np.mod(-rfbas[-1],Ls),), rfbas[1:]-rfbas[:-1]))
    
    timepos = np.cumsum(shift)
    nn = timepos[-1]
    timepos -= shift[0] # Calculate positions from shift vector
    
    wins = []
    for gii,tpii in zip(g, timepos):
        Lg = len(gii)
        win_range = np.arange(-(Lg//2)+tpii, Lg-(Lg//2)+tpii, dtype=int)
        win_range %= nn

#        Lg2 = Lg//2
#        oh = tpii
#        o = oh-Lg2
#        oe = oh+Lg2
#
#        if o < 0:
#            # wraparound is in first half
#            win_range = ((slice(o+nn,nn),slice(0,oh)),(slice(oh,oe),slice(0,0)))
#        elif oe > nn:
#            # wraparound is in second half
#            win_range = ((slice(o,oh),slice(0,0)),(slice(oh,nn),slice(0,oe-nn)))
#        else:
#            # no wraparound
#            win_range = ((slice(o,oh),slice(0,0)),(slice(oh,oe),slice(0,0)))

        wins.append(win_range)
        
    return wins,nn




# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)
"""

import numpy as np
cimport numpy as np

def nsgtf_loop(loopparams, np.ndarray ft not None, np.ndarray temp0 not None):
    cdef list c = [] # Initialization of the result
        
    # The actual transform
    # TODO: stuff loop into theano
    cdef int mii,Lg,col
    cdef np.ndarray gi1,gi2,t1,t2,ftw,temp,win_range
    for mii,_,gi1,gi2,win_range,Lg,col in loopparams:
#            Lg = len(gii)            
        # if the number of time channels is too small (mii < Lg), aliasing is introduced
        # wrap around and sum up in the end (below)
#            col = int(ceil(float(Lg)/mii)) # normally col == 1                        
#            assert col*mii >= Lg                        

        temp = temp0[:col*mii]

        # original version
#            t = ft[win_range]*N.fft.fftshift(N.conj(gii))
#            temp[:(Lg+1)//2] = t[Lg//2:]  # if mii is odd, this is of length mii-mii//2
#            temp[-(Lg//2):] = t[:Lg//2]  # if mii is odd, this is of length mii//2

        # modified version to avoid superfluous memory allocation
        t1 = temp[:(Lg+1)//2]
        t1[:] = gi1  # if mii is odd, this is of length mii-mii//2
        t2 = temp[-(Lg//2):]
        t2[:] = gi2  # if mii is odd, this is of length mii//2

        ftw = ft[win_range]
        t2 *= ftw[:Lg//2]
        t1 *= ftw[Lg//2:]

#            (wh1a,wh1b),(wh2a,wh2b) = win_range
#            t2[:wh1a.stop-wh1a.start] *= ft[wh1a]
#            t2[wh1a.stop-wh1a.start:] *= ft[wh1b]
#            t1[:wh2a.stop-wh2a.start] *= ft[wh2a]
#            t1[wh2a.stop-wh2a.start:] *= ft[wh2b]
        
        temp[(Lg+1)//2:-(Lg//2)] = 0  # clear gap (if any)
        
        if col > 1:
            temp = np.sum(temp.reshape((mii, -1)), axis=1)
        else:
            temp = temp.copy()

        c.append(temp)
    return c



import numpy as np
from nsgt.fft import rfftp, irfftp, fftp, ifftp
import unittest

class TestFFT(unittest.TestCase):
    def __init__(self, methodName, n=10000):
        super(TestFFT, self).__init__(methodName)
        self.n = n

    def test_rfft(self):
        seq = np.random.random(self.n)
        ft = rfftp()
        a = ft(seq)
        b = np.fft.rfft(seq)
        self.assertTrue(np.allclose(a, b))
        
    def test_irfft(self):
        seq = np.random.random(self.n)+np.random.random(self.n)*1.j
        outn = (self.n-1)*2 + np.random.randint(0,2) # even or odd output size
        ft = irfftp()
        a = ft(seq, outn=outn)
        b = np.fft.irfft(seq, n=outn)
        self.assertTrue(np.allclose(a, b))
        
    def test_fft(self):
        seq = np.random.random(self.n)
        ft = fftp()
        a = ft(seq)
        b = np.fft.fft(seq)
        self.assertTrue(np.allclose(a, b))
        
    def test_ifft(self):
        seq = np.random.random(self.n)+np.random.random(self.n)*1.j
        ft = ifftp()
        a = ft(seq)
        b = np.fft.ifft(seq)
        self.assertTrue(np.allclose(a, b))

def load_tests(*_):
    # seed random generators for unit testing
    np.random.seed(666)

    test_cases = unittest.TestSuite()
    for _ in range(100):
        l = int(10*np.random.uniform(2, 5))
        test_cases.addTest(TestFFT('test_rfft', l))
        test_cases.addTest(TestFFT('test_irfft', l))
        test_cases.addTest(TestFFT('test_fft', l))
        test_cases.addTest(TestFFT('test_ifft', l))
    return test_cases

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)
"""

import numpy as np


def reblock(sseq, blocksize, dtype=None, fulllast=True, padding=0, multichannel=False):
    block = None
    dt = None
    chns = None
    
    if multichannel:
        channelize = lambda s: s
        unchannelize = lambda s: s
    else:
        channelize = lambda s: (s,)
        unchannelize = lambda s: s[0]

    for si in sseq:
        # iterate through sequence of sequences

        si = channelize(si)
        
        while True:
            if block is None:
                if dt is None:
                    # output dtype still undefined
                    if dtype is None:
                        dt = type(si[0][0]) # take type is first input element
                    else:
                        dt = dtype
                chns = len(si)

                block = np.empty((chns,blocksize), dtype=dt)
                blockrem = block
                
            sout = [sj[:blockrem.shape[1]] for sj in si]
            avail = len(sout[0])
            for blr,souti in zip(blockrem, sout):
                blr[:avail] = souti # copy data per channel
            si = [sj[avail:] for sj in si]  # move ahead in input block
            blockrem = blockrem[:,avail:]  # move ahead in output block
            
            if blockrem.shape[1] == 0:
                # output block is full
                yield unchannelize(block)
                block = None
            if len(si[0]) == 0:
                # input block is exhausted
                break
            
    if block is not None:
        if fulllast:
            blockrem[:] = padding  # zero padding
            ret = block
        else:
            # output only filled part
            ret = block[:,:-len(blockrem[0])]
        yield unchannelize(ret)



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)


covered by Creative Commons Attribution-NonCommercial-ShareAlike license (CC BY-NC-SA)
http://creativecommons.org/licenses/by-nc-sa/3.0/at/deed.en


--
Original matlab code copyright follows:

AUTHOR(s) : Monika Dörfler, Gino Angelo Velasco, Nicki Holighaus, 2010-2011

COPYRIGHT : (c) NUHAG, Dept.Math., University of Vienna, AUSTRIA
http://nuhag.eu/
Permission is granted to modify and re-distribute this
code in any manner as long as this notice is preserved.
All standard disclaimers apply.

"""

__version__ = '0.18'

from .cq import NSGT, CQ_NSGT
from .slicq import NSGT_sliced, CQ_NSGT_sliced
from .fscale import Scale, OctScale, LogScale, LinScale, MelScale
from warnings import warn

try:
    from .audio import SndReader, SndWriter
except ImportError:
    warn("Audio IO routines (scikits.audio module) could not be imported")



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)
"""

import numpy as np
from math import ceil

from .util import chkM
from .fft import fftp, ifftp

try:
    # try to import cython version
    from _nsgtf_loop import nsgtf_loop
except ImportError:
    nsgtf_loop = None

if nsgtf_loop is None:
    from .nsgtf_loop import nsgtf_loop

if False:
    # what about theano?
    try:
        import theano as T
    except ImportError:
        T = None

try:
    import multiprocessing as MP
except ImportError:
    MP = None


#@profile
def nsgtf_sl(f_slices, g, wins, nn, M=None, real=False, reducedform=0, measurefft=False, multithreading=False):
    M = chkM(M,g)
    dtype = g[0].dtype
    
    fft = fftp(measure=measurefft, dtype=dtype)
    ifft = ifftp(measure=measurefft, dtype=dtype)
    
    if real:
        assert 0 <= reducedform <= 2
        sl = slice(reducedform,len(g)//2+1-reducedform)
    else:
        sl = slice(0,None)
    
    maxLg = max(int(ceil(float(len(gii))/mii))*mii for mii,gii in zip(M[sl],g[sl]))
    temp0 = None
    
    if multithreading and MP is not None:
        mmap = MP.Pool().map
    else:
        mmap = map

    loopparams = []
    for mii,gii,win_range in zip(M[sl],g[sl],wins[sl]):
        Lg = len(gii)
        col = int(ceil(float(Lg)/mii))
        assert col*mii >= Lg
        gi1 = gii[:(Lg+1)//2]
        gi2 = gii[-(Lg//2):]
        p = (mii,gii,gi1,gi2,win_range,Lg,col)
        loopparams.append(p)

    # main loop over slices
    for f in f_slices:
        Ls = len(f)
        
        # some preparation    
        ft = fft(f)

        if temp0 is None:
            # pre-allocate buffer (delayed because of dtype)
            temp0 = np.empty(maxLg, dtype=ft.dtype)
        
        # A small amount of zero-padding might be needed (e.g. for scale frames)
        if nn > Ls:
            ft = np.concatenate((ft, np.zeros(nn-Ls, dtype=ft.dtype)))
        
        # The actual transform
        c = nsgtf_loop(loopparams, ft, temp0)
            
        # TODO: if matrixform, perform "2D" FFT along one axis
        # this could also be nicely parallelized
        y = list(mmap(ifft,c))
        
        yield y

        
# non-sliced version
def nsgtf(f, g, wins, nn, M=None, real=False, reducedform=0, measurefft=False, multithreading=False):
    return next(nsgtf_sl((f,), g, wins, nn, M=M, real=real, reducedform=reducedform, measurefft=measurefft, multithreading=multithreading))



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)
"""

import numpy as np
from itertools import cycle, chain
from .util import hannwin

def slicequads(frec_sliced, hhop):
    slices = [[slice(hhop*((i+3-k*2)%4),hhop*((i+3-k*2)%4+1)) for i in range(4)] for k in range(2)]
    slices = cycle(slices)
    
    for fsl,sl in zip(frec_sliced, slices):
        yield [[fslc[sli] for fslc in fsl] for sli in sl]


def unslicing(frec_sliced, sl_len, tr_area, dtype=float, usewindow=True):
    hhop = sl_len//4    
    islices = slicequads(frec_sliced, hhop)
    
    if usewindow:
        tr_area2 = min(2*hhop-tr_area, 2*tr_area)
        htr = tr_area//2
        htr2 = tr_area2//2
        hw = hannwin(tr_area2)
        tw = np.zeros(sl_len, dtype=dtype)
        tw[max(hhop-htr-htr2, 0):hhop-htr] = hw[htr2:]
        tw[hhop-htr:3*hhop+htr] = 1
        tw[3*hhop+htr:min(3*hhop+htr+htr2, sl_len)] = hw[:htr2]
        tw = [tw[o:o+hhop] for o in range(0, sl_len, hhop)]
    else:
        tw = cycle((1,))
        
    # get first slice to deduce channels
    firstquad = next(islices)
    
    chns = len(firstquad[0]) # number of channels in first quad
    
    islices = chain((firstquad,), islices)
    
    output = [np.zeros((chns,hhop), dtype=dtype) for _ in range(4)]
    
    for quad in islices:
        for osl,isl,w in zip(output, quad, tw):
            # in a piecewise manner add slices to output stream 
            osl[:] += isl*w
        for _ in range(2):
            # absolutely first two should be padding (and discarded by the receiver)
            yield output.pop(0)
            output.append(np.zeros((chns,hhop), dtype=dtype))

    for _ in range(2):
        # absolutely last two should be padding (and discarded by the receiver)
        yield output.pop(0)

    # two more buffers remaining (and zero)



# -*- coding: utf-8

"""
Thomas Grill, 2011-2016
http://grrrr.org/nsgt

--
Original matlab code comments follow:

NSGFWIN.M
---------------------------------------------------------------
 [g,rfbas,M]=nsgfwin(fmin,bins,sr,Ls) creates a set of windows whose
 centers correspond to center frequencies to be
 used for the nonstationary Gabor transform with varying Q-factor. 
---------------------------------------------------------------

INPUT : fmin ...... Minimum frequency (in Hz)
        bins ...... Vector consisting of the number of bins per octave
        sr ........ Sampling rate (in Hz)
        Ls ........ Length of signal (in samples)

OUTPUT : g ......... Cell array of window functions.
         rfbas ..... Vector of positions of the center frequencies.
         M ......... Vector of lengths of the window functions.

AUTHOR(s) : Monika Dörfler, Gino Angelo Velasco, Nicki Holighaus, 2010

COPYRIGHT : (c) NUHAG, Dept.Math., University of Vienna, AUSTRIA
http://nuhag.eu/
Permission is granted to modify and re-distribute this
code in any manner as long as this notice is preserved.
All standard disclaimers apply.

EXTERNALS : firwin
"""

import numpy as np
from .util import hannwin, blackharr, blackharrcw
from math import ceil
from warnings import warn
from itertools import chain

def nsgfwin(f, q, sr, Ls, sliced=True, min_win=4, Qvar=1, dowarn=True, dtype=np.float64):
    nf = sr/2.

    lim = np.argmax(f > 0)
    if lim != 0:
        # f partly <= 0 
        f = f[lim:]
        q = q[lim:]
            
    lim = np.argmax(f >= nf)
    if lim != 0:
        # f partly >= nf 
        f = f[:lim]
        q = q[:lim]
    
    assert len(f) == len(q)
    assert np.all((f[1:]-f[:-1]) > 0)  # frequencies must be increasing
    assert np.all(q > 0)  # all q must be > 0
    
    qneeded = f*(Ls/(8.*sr))
    if np.any(q >= qneeded) and dowarn:
        warn("Q-factor too high for frequencies %s"%",".join("%.2f"%fi for fi in f[q >= qneeded]))
    
    fbas = f
    lbas = len(fbas)
    
    frqs = np.concatenate(((0.,),fbas,(nf,)))
    
    fbas = np.concatenate((frqs,sr-frqs[-2:0:-1]))

    # at this point: fbas.... frequencies in Hz
    
    fbas *= float(Ls)/sr
    
#    print "fbas",fbas
    
    # Omega[k] in the paper
    if sliced:
        M = np.zeros(fbas.shape, dtype=float)
        M[0] = 2*fbas[1]
        M[1] = fbas[1]/q[0] #(2**(1./bins[0])-2**(-1./bins[0]))
        for k in chain(range(2,lbas),(lbas+1,)):
            M[k] = fbas[k+1]-fbas[k-1]
        M[lbas] = fbas[lbas]/q[lbas-1] #(2**(1./bins[-1])-2**(-1./bins[-1]))
#        M[lbas+1] = fbas[lbas]/q[lbas-1] #(2**(1./bins[-1])-2**(-1./bins[-1]))
        M[lbas+2:2*(lbas+1)] = M[lbas:0:-1]
#        M[-1] = M[1]
        M *= Qvar/4.
        M = np.round(M).astype(int)
        M *= 4        
    else:
        M = np.zeros(fbas.shape, dtype=int)
        M[0] = np.round(2*fbas[1])
        for k in range(1,2*lbas+1):
            M[k] = np.round(fbas[k+1]-fbas[k-1])
        M[-1] = np.round(Ls-fbas[-2])
        
    np.clip(M, min_win, np.inf, out=M)

#    print "M",list(M)
    
    if sliced: 
        g = [blackharr(m).astype(dtype) for m in M]
    else:
        g = [hannwin(m).astype(dtype) for m in M]
    
    if sliced:
        for kk in (1,lbas+2):
            if M[kk-1] > M[kk]:
                g[kk-1] = np.ones(M[kk-1], dtype=g[kk-1].dtype)
                g[kk-1][M[kk-1]//2-M[kk]//2:M[kk-1]//2+int(ceil(M[kk]/2.))] = hannwin(M[kk])
        
        rfbas = np.round(fbas/2.).astype(int)*2
    else:
        fbas[lbas] = (fbas[lbas-1]+fbas[lbas+1])/2
        fbas[lbas+2] = Ls-fbas[lbas]
        rfbas = np.round(fbas).astype(int)
        
#    print "rfbas",rfbas
#    print "g",g

    return g,rfbas,M


def nsgfwin_new(f, q, sr, Ls, sliced=True, min_win=4, Qvar=1, dowarn=True):
    nf = sr/2.

    lim = np.argmax(f > 0)
    if lim != 0:
        # f partly <= 0 
        f = f[lim:]
        q = q[lim:]
            
    lim = np.argmax(f >= nf)
    if lim != 0:
        # f partly >= nf 
        f = f[:lim]
        q = q[:lim]
    
    assert len(f) == len(q)
    assert np.all((f[1:]-f[:-1]) > 0)  # frequencies must be increasing
    assert np.all(q > 0)  # all q must be > 0
    
    qneeded = f*(Ls/(8.*sr))
    if np.any(q >= qneeded) and dowarn:
        warn("Q-factor too high for frequencies %s"%",".join("%.2f"%fi for fi in f[q >= qneeded]))
    
    fbas = f
    lbas = len(fbas)
    
    frqs = np.concatenate(((0.,),fbas,(nf,)))
    
    fbas = np.concatenate((frqs,sr-frqs[-2:0:-1]))

    # at this point: fbas.... frequencies in Hz
    
    fbas *= float(Ls)/sr
    
#    print "fbas",fbas
    
    # Omega[k] in the paper
    if sliced:
        M = np.zeros(fbas.shape, dtype=float)
        M[0] = 2*fbas[1]
        M[1] = fbas[1]/q[0] #(2**(1./bins[0])-2**(-1./bins[0]))
        for k in chain(range(2,lbas),(lbas+1,)):
            M[k] = fbas[k+1]-fbas[k-1]
        M[lbas] = fbas[lbas]/q[lbas-1] #(2**(1./bins[-1])-2**(-1./bins[-1]))
#        M[lbas+1] = fbas[lbas]/q[lbas-1] #(2**(1./bins[-1])-2**(-1./bins[-1]))
        M[lbas+2:2*(lbas+1)] = M[lbas:0:-1]
#        M[-1] = M[1]

###        M *= Qvar/4.
###        M = N.round(M).astype(int)
###        M *= 4        
    else:
        M = np.zeros(fbas.shape, dtype=int)
        M[0] = np.round(2*fbas[1])
        for k in range(1, 2*lbas+1):
            M[k] = np.round(fbas[k+1]-fbas[k-1])
        M[-1] = np.round(Ls-fbas[-2])
        
    np.clip(M, min_win, np.inf, out=M)

    if sliced: 
###        g = [blackharr(m) for m in M]

        rfbas = np.concatenate((np.floor(fbas[:lbas+2]), np.ceil(fbas[lbas+2:])))
        corr_shift = fbas-rfbas

#        shift = N.concatenate(([(-rfbas[-1])%Ls],N.diff(rfbas)))

        g,M = np.array([blackharrcw(m*Qvar, csh) for m,csh in zip(M, corr_shift)]).T  ####
#        M = N.ceil(M/4).astype(int)*4  # bailing out for some strange reason....
        M = np.array([ceil(m/4)*4 for m in M], dtype=int)
    else:
        g = [hannwin(m) for m in M]
    
    
    
    if sliced:
        for kk in (1,lbas+2):
            if M[kk-1] > M[kk]:
                g[kk-1] = np.ones(M[kk-1],dtype=g[kk-1].dtype)
                g[kk-1][M[kk-1]//2-M[kk]//2:M[kk-1]//2+ceil(M[kk]/2.)] = hannwin(M[kk])
        
        rfbas = np.round(fbas/2.).astype(int)*2
    else:
        fbas[lbas] = (fbas[lbas-1]+fbas[lbas+1])/2
        fbas[lbas+2] = Ls-fbas[lbas]
        rfbas = np.round(fbas).astype(int)
        
#    print "rfbas",rfbas
#    print "g",g    
    
    return g,rfbas,M



# -*- coding: utf-8

"""
Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)

--
Original matlab code comments follow:

NSDUAL.M - Nicki Holighaus 02.02.11

Computes (for the painless case) the dual frame corresponding to a given 
non-stationary Gabor frame specified by the windows 'g' and time shifts
'shift'.

Note, the time shifts corresponding to the dual window sequence is the
same as the original shift sequence and as such already given.

This routine's output can be used to achieve reconstruction of a signal 
from its non-stationary Gabor coefficients using the inverse 
non-stationary Gabor transform 'nsigt'.

More information on Non-stationary Gabor transforms
can be found at:

http://www.univie.ac.at/nonstatgab/

minor edit by Gino Velasco 23.02.11
"""

import numpy as np

from .util import chkM

def nsdual(g, wins, nn, M=None):

    M = chkM(M,g)

    # Construct the diagonal of the frame operator matrix explicitly
    x = np.zeros((nn,), dtype=float)
    for gi,mii,sl in zip(g, M, wins):
        xa = np.square(np.fft.fftshift(gi))
        xa *= mii
        x[sl] += xa

        # could be more elegant...
#        (w1a,w1b),(w2a,w2b) = sl
#        x[w1a] += xa[:w1a.stop-w1a.start]
#        xa = xa[w1a.stop-w1a.start:]
#        x[w1b] += xa[:w1b.stop-w1b.start]
#        xa = xa[w1b.stop-w1b.start:]
#        x[w2a] += xa[:w2a.stop-w2a.start]
#        xa = xa[w2a.stop-w2a.start:]
#        x[w2b] += xa[:w2b.stop-w2b.start]
##        xa = xa[w1b.stop-w1b.start:]

    # Using the frame operator and the original window sequence, compute 
    # the dual window sequence
#    gd = [gi/N.fft.ifftshift(N.hstack((x[wi[0][0]],x[wi[0][1]],x[wi[1][0]],x[wi[1][1]]))) for gi,wi in izip(g,wins)]
    gd = [gi/np.fft.ifftshift(x[wi]) for gi,wi in zip(g,wins)]
    return gd



# -*- coding: utf-8

"""
Thomas Grill, 2011-2015
http://grrrr.org/nsgt

--
Original matlab code comments follow:

NSGFWIN.M
---------------------------------------------------------------
 [g,rfbas,M]=nsgfwin(fmin,bins,sr,Ls) creates a set of windows whose
 centers correspond to center frequencies to be
 used for the nonstationary Gabor transform with varying Q-factor. 
---------------------------------------------------------------

INPUT : fmin ...... Minimum frequency (in Hz)
        bins ...... Vector consisting of the number of bins per octave
        sr ........ Sampling rate (in Hz)
        Ls ........ Length of signal (in samples)

OUTPUT : g ......... Cell array of window functions.
         rfbas ..... Vector of positions of the center frequencies.
         M ......... Vector of lengths of the window functions.

AUTHOR(s) : Monika Dörfler, Gino Angelo Velasco, Nicki Holighaus, 2010

COPYRIGHT : (c) NUHAG, Dept.Math., University of Vienna, AUSTRIA
http://nuhag.eu/
Permission is granted to modify and re-distribute this
code in any manner as long as this notice is preserved.
All standard disclaimers apply.

EXTERNALS : firwin
"""

import numpy as np
from .util import hannwin,_isseq

def nsgfwin(fmin, fmax ,bins, sr, Ls, min_win=4):

    nf = sr/2
    
    if fmax > nf:
        fmax = nf
    
    b = np.ceil(np.log2(fmax/fmin))+1

    if not _isseq(bins):
        bins = np.ones(b,dtype=int)*bins
    elif len(bins) < b:
        # TODO: test this branch!
        bins[bins <= 0] = 1
        bins = np.concatenate((bins, np.ones(b-len(bins), dtype=int)*np.min(bins)))
    
    fbas = []
    for kk,bkk in enumerate(bins):
        r = np.arange(kk*bkk, (kk+1)*bkk, dtype=float)
        # TODO: use N.logspace instead
        fbas.append(2**(r/bkk)*fmin)
    fbas = np.concatenate(fbas)

    if fbas[np.min(np.where(fbas>=fmax))] >= nf:
        fbas = fbas[:np.max(np.where(fbas<fmax))+1]
    else:
        # TODO: test this branch!
        fbas = fbas[:np.min(np.where(fbas>=fmax))+1]
    
    lbas = len(fbas)
    fbas = np.concatenate(((0.,), fbas, (nf,), sr-fbas[::-1]))
    fbas *= float(Ls)/sr
    
    # TODO: put together with array indexing
    M = np.empty(fbas.shape, dtype=int)
    M[0] = np.round(2.*fmin*Ls/sr)
    for k in range(1, 2*lbas+1):
        M[k] = np.round(fbas[k+1]-fbas[k-1])
    M[-1] = np.round(Ls-fbas[-2])
    
    M = np.clip(M, min_win, np.inf).astype(int)
    g = [hannwin(m) for m in M]
    
    fbas[lbas] = (fbas[lbas-1]+fbas[lbas+1])/2
    fbas[lbas+2] = Ls-fbas[lbas]
    rfbas = np.round(fbas).astype(int)
    
    return g,rfbas,M



import numpy as np
from nsgt import NSGT_sliced, OctScale
import unittest

# reproducible noise signal
maxlen = 50000
np.random.seed(666)
rndsig = np.random.random((maxlen,))
#assert np.allclose(rndsig[:3],[ 0.70043712, 0.84418664, 0.67651434])

class TestNSGT_slices(unittest.TestCase):

    def runit(self, siglen, fmin, fmax, obins, sllen, trlen, real):
        sig = rndsig[:siglen]

        scale = OctScale(fmin, fmax, obins)
        nsgt = NSGT_sliced(scale, fs=44100, sl_len=sllen, tr_area=trlen, real=real)

        c = nsgt.forward((sig,))

        rc = nsgt.backward(c)

        s_r = np.concatenate(list(map(list,rc)))[:len(sig)]
        
        close = np.allclose(sig, s_r, atol=1.e-3)
        if not close:
            print("Failing params:", siglen, fmin, fmax, obins, sllen, trlen, real)
            dev = np.abs(s_r-sig)
            print("Error", np.where(dev>1.e-3), np.max(dev))
        self.assertTrue(close)


    def test_1d1(self):
        self.runit(*list(map(int,"100000 100 18200 2 20000 5000 1".split())))
        
    def test_1d11(self):
        self.runit(*list(map(int,"100000 80 18200 6 20000 5000 1".split())))
        
    def test_1(self):
        self.runit(*list(map(int,"100000 99 19895 6 84348 5928 1".split())))
        
    def test_1a(self):
        self.runit(*list(map(int,"100000 99 19895 6 84348 5928 0".split())))
        
    def test_1b(self):
        self.runit(*list(map(int,"100000 100 20000 6 80000 5000 1".split())))
        
    def test_1c(self):
        self.runit(*list(map(int,"100000 100 19000 6 80000 5000 1".split())))
        
    def test_1d2(self):
        self.runit(*list(map(int,"100000 100 18100 6 20000 5000 1".split())))
        
    def test_1e(self):
        self.runit(*list(map(int,"100000 100 18000 6 20000 5000 1".split())))
        
    def test_err1(self):
        self.runit(*list(map(int,"30549 104 11970 25 7286 2030 1".split())))
        
    def test_err2(self):
        self.runit(*list(map(int,"19746 88 19991 21 12674 4030 0".split())))
        
    def test_err3(self):
        self.runit(*list(map(int,"92507 114 18387 20 29306 11848 1".split())))
        
    def test_err4(self):
        self.runit(*list(map(int,"20724 191 2843 16 22354 6590 1".split())))
        
    def test_err5(self):
        self.runit(*list(map(int,"10712 97 19363 3 10238 1876 1".split())))
        
    def test_err6(self):
        self.runit(*list(map(int,"262597 100 15786 16 2858 556 1".split())))
        
    def gtest_oct(self):
        siglen = int(10**np.random.uniform(4,6))
        fmin = np.random.randint(200)+80
        fmax = max(np.random.randint(22048-fmin)+fmin,10000)
        obins = np.random.randint(24)+1
        sllen = max(1000,np.random.randint(maxlen))//4*4
        trlen = max(2,np.random.randint(sllen//2-2))//2*2
        real = np.random.randint(2)
        self.runit(siglen, fmin, fmax, obins, sllen, trlen, real)

def load_tests(*_):
    test_cases = unittest.TestSuite()
#        for t in 'test_1d1 test_1d11 test_1 test_1a test_1b test_1c test_1d2 test_1e test_err1 test_err2 test_err3 test_err4 test_err5 test_err6'.split():
#            test_cases.addTest(TestNSGT_slices(t))
    
    for _ in range(50):
        test_cases.addTest(TestNSGT_slices('gtest_oct'))
    return test_cases

if __name__ == '__main__':
    unittest.main()



import numpy as np
from nsgt import CQ_NSGT
import unittest

class Test_CQ_NSGT(unittest.TestCase):

    def test_transform(self, length=100000, fmin=50, fmax=22050, bins=12, fs=44100):
        s = np.random.random(length)
        nsgt = CQ_NSGT(fmin, fmax, bins, fs, length)
        
        # forward transform 
        c = nsgt.forward(s)
        # inverse transform 
        s_r = nsgt.backward(c)
        
        self.assertTrue(np.allclose(s, s_r, atol=1e-07))

def load_tests(*_):
    # seed random generators for unit testing
    np.random.seed(666)

    test_cases = unittest.TestSuite()
    for _ in range(100):
        test_cases.addTest(Test_CQ_NSGT('test_transform'))
    return test_cases

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)
"""

import numpy as np
from warnings import warn

# Try engines in order of:
# PyFFTW3
# pyFFTW
# numpy.fftpack
try:
    import fftw3, fftw3f
except ImportError:
    fftw3 = None
    fftw3f = None
    try:
        import pyfftw
    except ImportError:
        pyfftw = None


if fftw3 is not None and fftw3f is not None:
    ENGINE = "PYFFTW3"
    # Use fftw3 methods
    class fftpool:
        def __init__(self, measure, dtype=float):
            self.measure = measure
            self.dtype = np.dtype(dtype)
            dtsz = self.dtype.itemsize
            if dtsz == 4:
                self.tpfloat = np.float32
                self.tpcplx = np.complex64
                self.fftw = fftw3f
            elif dtsz == 8:
                self.tpfloat = np.float64
                self.tpcplx = np.complex128
                self.fftw = fftw3
            else:
                raise TypeError("nsgt.fftpool: dtype '%s' not supported"%repr(self.dtype))
            self.pool = {}

        def __call__(self, x, outn=None, ref=False):
            lx = len(x)
            try:
                transform = self.pool[lx]
            except KeyError:
                transform = self.init(lx, measure=self.measure, outn=outn)
                self.pool[lx] = transform
            plan,pre,post = transform
            if pre is not None:
                x = pre(x)
            plan.inarray[:] = x
            plan()
            if not ref:
                tx = plan.outarray.copy()
            else:
                tx = plan.outarray
            if post is not None:
                tx = post(tx)
            return tx

    class fftp(fftpool):
        def __init__(self, measure=False, dtype=float):
            fftpool.__init__(self, measure, dtype=dtype)
        def init(self, n, measure, outn):
            inp = self.fftw.create_aligned_array(n, dtype=self.tpcplx)
            outp = self.fftw.create_aligned_array(n, dtype=self.tpcplx)
            plan = self.fftw.Plan(inp, outp, direction='forward', flags=('measure' if measure else 'estimate',))
            return (plan,None,None)

    class rfftp(fftpool):
        def __init__(self, measure=False, dtype=float):
            fftpool.__init__(self, measure, dtype=dtype)
        def init(self, n, measure, outn):
            inp = self.fftw.create_aligned_array(n, dtype=self.tpfloat)
            outp = self.fftw.create_aligned_array(n//2+1, dtype=self.tpcplx)
            plan = self.fftw.Plan(inp, outp, direction='forward', realtypes='halfcomplex r2c',flags=('measure' if measure else 'estimate',))
            return (plan,None,None)

    class ifftp(fftpool):
        def __init__(self, measure=False, dtype=float):
            fftpool.__init__(self, measure, dtype=dtype)
        def init(self, n, measure, outn):
            inp = self.fftw.create_aligned_array(n, dtype=self.tpcplx)
            outp = self.fftw.create_aligned_array(n, dtype=self.tpcplx)
            plan = self.fftw.Plan(inp, outp, direction='backward', flags=('measure' if measure else 'estimate',))
            return (plan,None,lambda x: x/len(x))

    class irfftp(fftpool):
        def __init__(self, measure=False, dtype=float):
            fftpool.__init__(self, measure, dtype=dtype)
        def init(self, n, measure, outn):
            inp = self.fftw.create_aligned_array(n, dtype=self.tpcplx)
            outp = self.fftw.create_aligned_array(outn if outn is not None else (n-1)*2, dtype=self.tpfloat)
            plan = self.fftw.Plan(inp, outp, direction='backward', realtypes='halfcomplex c2r', flags=('measure' if measure else 'estimate',))
            return (plan,lambda x: x[:n],lambda x: x/len(x))
elif pyfftw is not None:
    ENGINE = "PYFFTW"
    # Monkey patch in pyFFTW Numpy interface
    np.fft = pyfftw.interfaces.numpy_fft
    original_fft = np.fft
    # Enable cache to keep wisdom, etc.
    pyfftw.interfaces.cache.enable()
else:
    # fall back to numpy methods
    warn("nsgt.fft falling back to numpy.fft")
    ENGINE = "NUMPY"

if ENGINE in ["PYFFTW", "NUMPY"]:
    class fftp:
        def __init__(self, measure=False, dtype=float):
            pass
        def __call__(self,x, outn=None, ref=False):
            return np.fft.fft(x)
    class ifftp:
        def __init__(self, measure=False, dtype=float):
            pass
        def __call__(self,x, outn=None, n=None, ref=False):
            return np.fft.ifft(x,n=n)
    class rfftp:
        def __init__(self, measure=False, dtype=float):
            pass
        def __call__(self,x, outn=None, ref=False):
            return np.fft.rfft(x)
    class irfftp:
        def __init__(self, measure=False, dtype=float):
            pass
        def __call__(self,x,outn=None,ref=False):
            return np.fft.irfft(x,n=outn)



#! /usr/bin/env python
# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2020
http://grrrr.org/nsgt
"""

import numpy as np
from nsgt import NSGT, LogScale, LinScale, MelScale, OctScale

try:
    from scikits.audiolab import Sndfile, Format
except:
    Sndfile = None
    
if Sndfile is None:
    from pysndfile import PySndfile, construct_format
    
import os


def cputime():
    utime, stime, cutime, cstime, elapsed_time = os.times()
    return utime

class interpolate:
    def __init__(self,cqt,Ls):
        from scipy.interpolate import interp1d
        self.intp = [interp1d(np.linspace(0, Ls, len(r)), r) for r in cqt]
    def __call__(self,x):
        try:
            len(x)
        except:
            return np.array([i(x) for i in self.intp])
        else:
            return np.array([[i(xi) for i in self.intp] for xi in x])


from argparse import ArgumentParser
parser = ArgumentParser()

parser.add_argument("input", type=str, help="input audio file")
parser.add_argument('-o', "--output", type=str, help="output audio file")
parser.add_argument('-f', "--fmin", type=float, default=80, help="minimum frequency (default=%(default)s)")
parser.add_argument('-F', "--fmax", type=float, default=22050, help="maximum frequency (default=%(default)s)")
parser.add_argument('-s', "--scale", choices=("oct","log","mel"), default='oct', help="frequency scale (oct,log,mel)")
parser.add_argument('-b', "--bins", type=int, default=24, help="frequency bins (total or per octave, default=%(default)s)")
parser.add_argument('-r', "--real", action='store_true', help="assume real signal")
parser.add_argument('-m', "--matrixform", action='store_true', help="use regular time division (matrix form)")
parser.add_argument('-l', "--reducedform", action='count', default=0, help="if real==1: omit bins for f=0 and f=fs/2 (lossy=1), or also the transition bands (lossy=2)")
parser.add_argument('-t', "--time", type=int, default=1, help="timing calculation n-fold (default=%(default)s)")
parser.add_argument('-p', "--plot", action='store_true', help="plot results (needs installed matplotlib and scipy packages)")

args = parser.parse_args()
if not os.path.exists(args.input):
    parser.error("Input file '%s' not found"%args.input)

# Read audio data
if Sndfile is not None:
    sf = Sndfile(args.input)
    fs = sf.samplerate
    samples = sf.nframes
else:
    sf = PySndfile(args.input)
    fs = sf.samplerate()
    samples = sf.frames()
s = sf.read_frames(samples)

if s.ndim > 1: 
    s = np.mean(s, axis=1)
    
scales = {'log':LogScale, 'lin':LinScale, 'mel':MelScale, 'oct':OctScale}
try:
    scale = scales[args.scale]
except KeyError:
    parser.error('scale unknown')

scl = scale(args.fmin, args.fmax, args.bins)

times = []

for _ in range(args.time or 1):
    t1 = cputime()
    
    # calculate transform parameters
    Ls = len(s)
    
    nsgt = NSGT(scl, fs, Ls, real=args.real, matrixform=args.matrixform, reducedform=args.reducedform)
    
    # forward transform 
    c = nsgt.forward(s)

    c = np.asarray(c)
    print("Coefficients:", c.shape)
#        print "c",len(c),N.array(map(len,c))

    # inverse transform 
    s_r = nsgt.backward(c)

    t2 = cputime()
    times.append(t2-t1)

norm = lambda x: np.sqrt(np.sum(np.abs(np.square(x))))
rec_err = norm(s-s_r)/norm(s)
print("Reconstruction error: %.3e"%rec_err)
print("Calculation time: %.3f±%.3fs (min=%.3f s)"%(np.mean(times), np.std(times)/2, np.min(times)))

if args.output:
    print("Writing audio file '%s'"%args.output)
    if Sndfile is not None:
        sf = Sndfile(args.output, mode='w', format=Format('wav','pcm24'), channels=1, samplerate=fs)
    else:
        sf = PySndfile(args.output, mode='w', format=construct_format('wav','pcm24'), channels=1, samplerate=fs)
    sf.write_frames(s_r)
    sf.close()
    print("Done")

if args.plot:
    print("Preparing plot")
    import matplotlib.pyplot as pl
    # interpolate CQT to get a grid
    hf = -1 if args.real else len(c)//2
    if not args.matrixform:
        x = np.linspace(0, Ls, 2000)
        grid = interpolate(map(np.abs, c[2:hf]), Ls)(x).T
    else:
        grid = np.abs(c[2:hf])
    np.log10(grid, out=grid)
    grid *= 20
    pmax = np.percentile(grid, 99.99)
    pl.imshow(grid, aspect='auto', origin='lower', vmin=pmax-80, vmax=pmax)
    pl.colorbar()
    pl.show()



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)
"""

import numpy as np
cimport numpy as np
from itertools import izip

def nsigtf_loop(loopparams, np.ndarray fr not None, fc):
    fr[:] = 0.
    # The overlap-add procedure including multiplication with the synthesis windows
    # TODO: stuff loop into theano
    cdef np.ndarray gdii,t,temp,t1,t2,wr1,wr2
    cdef slice sl1,sl2
    for t,(gdii,wr1,wr2,sl1,sl2,temp) in izip(fc, loopparams):
        t1 = temp[sl1]
        t2 = temp[sl2]
        t1[:] = t[sl1]
        t2[:] = t[sl2]
        temp *= gdii
        temp *= len(t)

        fr[wr1] += t2
        fr[wr2] += t1

    return fr



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2021
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)
"""

import numpy as np
import subprocess as sp
import os.path
import re
import sys
from functools import reduce

try:
    from pysndfile import PySndfile, construct_format
except:
    PySndfile = None
    
def sndreader(sf, blksz=2**16, dtype=np.float32):
    frames = sf.frames()
    if dtype is float:
        dtype = np.float64 # scikits.audiolab needs numpy types
    if blksz < 0:
        blksz = frames
    if sf.channels() > 1: 
        channels = lambda s: s.T
    else:
        channels = lambda s: s.reshape((1,-1))
    for offs in range(0, frames, blksz):
        data = sf.read_frames(min(frames-offs, blksz), dtype=dtype)
        yield channels(data)
    
def sndwriter(sf, blkseq, maxframes=None):
    written = 0
    for b in blkseq:
        b = b.T
        if maxframes is not None: 
            b = b[:maxframes-written]
        sf.write_frames(b)
        written += len(b)

def findfile(fn, path=os.environ['PATH'].split(os.pathsep), matchFunc=os.path.isfile):
    for dirname in path:
        candidate = os.path.join(dirname, fn)
        if matchFunc(candidate):
            return candidate
    return None


class SndReader:
    def __init__(self, fn, sr=None, chns=None, blksz=2**16, dtype=np.float32):
        fnd = False
                
        if not fnd and (PySndfile is not None):
            try:
                sf = PySndfile(fn, mode='r')
            except IOError:
                pass
            else:
                if (sr is None or sr == sf.samplerate()) and (chns is None or chns == sf.channels()):
                    # no resampling required
                    self.channels = sf.channels()
                    self.samplerate = sf.samplerate()
                    self.frames = sf.frames()
                
                    self.rdr = sndreader(sf, blksz, dtype=dtype)
                    fnd = True
        
        if not fnd:
            ffmpeg = findfile('ffmpeg') or findfile('avconv')
            if ffmpeg is not None:
                pipe = sp.Popen([ffmpeg,'-i', fn,'-'],stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE)
                fmtout = pipe.stderr.read()
                if (sys.version_info > (3, 0)):
                    fmtout = fmtout.decode()
                m = re.match(r"^(ffmpeg|avconv) version.*Duration: (\d\d:\d\d:\d\d.\d\d),.*Audio: (.+), (\d+) Hz, (.+), (.+), (\d+) kb/s", " ".join(fmtout.split('\n')))
                if m is not None:
                    self.samplerate = int(m.group(4)) if not sr else int(sr)
                    chdef = m.group(5)
                    if chdef.endswith(" channels") and len(chdef.split()) == 2:
                        self.channels = int(chdef.split()[0])
                    else:
                        try:
                            self.channels = {'mono':1, '1 channels (FL+FR)':1, 'stereo':2, 'hexadecagonal':16}[chdef] if not chns else chns
                        except:
                            print(f"Channel definition '{chdef}' unknown")
                            raise
                    dur = reduce(lambda x,y: x*60+y, list(map(float, m.group(2).split(':'))))
                    self.frames = int(dur*self.samplerate)  # that's actually an estimation, because of potential resampling with round-off errors
                    pipe = sp.Popen([ffmpeg,
                        '-i', fn,
                        '-f', 'f32le',
                        '-acodec', 'pcm_f32le',
                        '-ar', str(self.samplerate),
                        '-ac', str(self.channels),
                        '-'],
    #                    bufsize=self.samplerate*self.channels*4*50,
                        stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE)
                    def rdr():
                        bufsz = (blksz//self.channels)*self.channels*4
                        while True:
                            data = pipe.stdout.read(bufsz)
                            if len(data) == 0:
                                break
                            data = np.fromstring(data, dtype=dtype)
                            yield data.reshape((-1, self.channels)).T
                    self.rdr = rdr()
                    fnd = True
                
        if not fnd:
            raise IOError("Format not usable")
        
    def __call__(self):
        return self.rdr


class SndWriter:
    def __init__(self, fn, samplerate, filefmt='wav', datafmt='pcm16', channels=1):
        fmt = construct_format(filefmt, datafmt)
        self.sf = PySndfile(fn, mode='w', format=fmt, channels=channels, samplerate=samplerate)
        
    def __call__(self, sigblks, maxframes=None):
        sndwriter(self.sf, sigblks, maxframes=None)




# -*- coding: utf-8

"""
Thomas Grill, 2011-2015
http://grrrr.org/nsgt

--
Original matlab code comments follow:

NSIGTF.N - Gino Velasco 24.02.11

fr = nsigtf(c,gd,shift,Ls)

This is a modified version of nsigt.m for the case where the resolution 
evolves over frequency.

Given the cell array 'c' of non-stationary Gabor coefficients, and a set 
of windows and frequency shifts, this function computes the corresponding 
inverse non-stationary Gabor transform.

Input: 
          c           : Cell array of non-stationary Gabor coefficients
          gd          : Cell array of Fourier transforms of the synthesis 
                        windows
          shift       : Vector of frequency shifts
          Ls          : Length of the analyzed signal

Output:
          fr          : Synthesized signal

If a non-stationary Gabor frame was used to produce the coefficients 
and 'gd' is a corresponding dual frame, this function should give perfect 
reconstruction of the analyzed signal (up to numerical errors).

The inverse transform is computed by simple 
overlap-add. For each entry of the cell array c,
the coefficients of frequencies around a certain 
position in time, the Fourier transform
is taken, giving 'frequency slices' of a signal.
These slices are added onto each other with an overlap
depending on the window lengths and positions, thus
(re-)constructing the frequency side signal. In the
last step, an inverse Fourier transform brings the signal
back to the time side.

More information can be found at:
http://www.univie.ac.at/nonstatgab/

Edited by Nicki Holighaus 01.03.11
"""

import numpy as np
from itertools import chain
from .fft import fftp, ifftp, irfftp

try:
    # try to import cython version
    from _nsigtf_loop import nsigtf_loop
except ImportError:
    nsigtf_loop = None

if nsigtf_loop is None:
    from .nsigtf_loop import nsigtf_loop

if False:
    # what about theano?
    try:
        import theano as T
    except ImportError:
        T = None
    
try:
    import multiprocessing as MP
except ImportError:
    MP = None
    

#@profile
def nsigtf_sl(cseq, gd, wins, nn, Ls=None, real=False, reducedform=0, measurefft=False, multithreading=False):
    cseq = iter(cseq)
    dtype = gd[0].dtype

    fft = fftp(measure=measurefft, dtype=dtype)
    ifft = irfftp(measure=measurefft, dtype=dtype) if real else ifftp(measure=measurefft, dtype=dtype)
    
    if real:
        ln = len(gd)//2+1-reducedform*2
        fftsymm = lambda c: np.hstack((c[0],c[-1:0:-1])).conj()
        if reducedform:
            # no coefficients for f=0 and f=fs/2
            def symm(_fc):
                fc = list(_fc)
                return chain(fc, map(fftsymm, fc[::-1]))
            sl = lambda x: chain(x[reducedform:len(gd)//2+1-reducedform],x[len(gd)//2+reducedform:len(gd)+1-reducedform])
        else:
            def symm(_fc):
                fc = list(_fc)
                return chain(fc, map(fftsymm, fc[-2:0:-1]))
            sl = lambda x: x
    else:
        ln = len(gd)
        symm = lambda fc: fc
        sl = lambda x: x
        
    maxLg = max(len(gdii) for gdii in sl(gd))

    # get first slice
    c0 = next(cseq)

    fr = np.empty(nn, dtype=c0[0].dtype)  # Allocate output
    temp0 = np.empty(maxLg, dtype=fr.dtype)  # pre-allocation
    
    if multithreading and MP is not None:
        mmap = MP.Pool().map
    else:
        mmap = map

    loopparams = []
    for gdii,win_range in zip(sl(gd), sl(wins)):
        Lg = len(gdii)
        temp = temp0[:Lg]
        wr1 = win_range[:(Lg)//2]
        wr2 = win_range[-((Lg+1)//2):]
#        wr1,wr2 = win_range
        sl1 = slice(None, (Lg+1)//2)
        sl2 = slice(-(Lg//2), None)
        p = (gdii,wr1,wr2,sl1,sl2,temp)
        loopparams.append(p)
        
    # main loop over slices
    for c in chain((c0,),cseq):
        assert len(c) == ln

        # do transforms on coefficients
        # TODO: for matrixform we could do a FFT on the whole matrix along one axis
        # this could also be nicely parallalized
        fc = mmap(fft, c)
        fc = symm(fc)
        
        # The overlap-add procedure including multiplication with the synthesis windows
        fr = nsigtf_loop(loopparams, fr, fc)

        ftr = fr[:nn//2+1] if real else fr

        sig = ifft(ftr, outn=nn)

        sig = sig[:Ls] # Truncate the signal to original length (if given)

        yield sig

# non-sliced version
def nsigtf(c, gd, wins, nn, Ls=None, real=False, reducedform=0, measurefft=False, multithreading=False):
    return next(nsigtf_sl((c,), gd, wins, nn, Ls=Ls, real=real, reducedform=reducedform, measurefft=measurefft, multithreading=multithreading))



#! /usr/bin/env python 
# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2016
http://grrrr.org/nsgt
"""

import numpy as np
import os
from warnings import warn

from nsgt import NSGT_sliced, LogScale, LinScale, MelScale, OctScale, SndReader


def assemble_coeffs(cqt, ncoefs):
    """
    Build a sequence of blocks out of incoming overlapping CQT slices
    """
    cqt = iter(cqt)
    cqt0 = next(cqt)
    cq0 = np.asarray(cqt0).T
    shh = cq0.shape[0]//2
    out = np.empty((ncoefs, cq0.shape[1], cq0.shape[2]), dtype=cq0.dtype)
    
    fr = 0
    sh = max(0, min(shh, ncoefs-fr))
    out[fr:fr+sh] = cq0[sh:] # store second half
    # add up slices
    for cqi in cqt:
        cqi = np.asarray(cqi).T
        out[fr:fr+sh] += cqi[:sh]
        cqi = cqi[sh:]
        fr += sh
        sh = max(0, min(shh, ncoefs-fr))
        out[fr:fr+sh] = cqi[:sh]
        
    return out[:fr]


from argparse import ArgumentParser
parser = ArgumentParser()

parser.add_argument("input", type=str, help="Input file")
parser.add_argument("--output", type=str, help="Output data file (.npz, .hd5, .pkl)")
parser.add_argument("--sr", type=int, default=44100, help="Sample rate used for the NSGT (default=%(default)s)")
parser.add_argument("--data-times", type=str, default='times', help="Data name for times (default='%(default)s')")
parser.add_argument("--data-frqs", type=str, default='f', help="Data name for frequencies (default='%(default)s')")
parser.add_argument("--data-qs", type=str, default='q', help="Data name for q factors (default='%(default)s')")
parser.add_argument("--data-coefs", type=str, default='coefs', help="Data name for coefficients (default='%(default)s')")
parser.add_argument("--fps", type=float, default=0, help="Approx. time resolution for features in fps (default=%(default)s)")
parser.add_argument("--fps-pooling", choices=('max','mean','median'), default='max', help="Temporal pooling function for features (default='%(default)s')")
parser.add_argument("--fmin", type=float, default=50, help="Minimum frequency in Hz (default=%(default)s)")
parser.add_argument("--fmax", type=float, default=22050, help="Maximum frequency in Hz (default=%(default)s)")
parser.add_argument("--scale", choices=('oct','log','mel'), default='log', help="Frequency scale oct, log, lin, or mel (default='%(default)s')")
parser.add_argument("--bins", type=int, default=50, help="Number of frequency bins (total or per octave, default=%(default)s)")
parser.add_argument("--mag-scale", choices=('dB','log'), default='dB', help="Magnitude scale dB or log (default='%(default)s')")
parser.add_argument("--sllen", type=int, default=2**20, help="Slice length in samples (default=%(default)s)")
parser.add_argument("--trlen", type=int, default=2**18, help="Transition area in samples (default=%(default)s)")
parser.add_argument("--real", action='store_true', help="Assume real signal")
parser.add_argument("--matrixform", action='store_true', help="Use regular time division over frequency bins (matrix form)")
parser.add_argument("--reducedform", type=int, default=0, help="If real, omit bins for f=0 and f=fs/2 (--reducedform=1), or also the transition bands (--reducedform=2) (default=%(default)s)")
parser.add_argument("--recwnd", action='store_true', help="Use reconstruction window")
parser.add_argument("--multithreading", action='store_true', help="Use multithreading")
parser.add_argument("--downmix-after", action='store_true', help="Downmix signal after spectrogram generation")
parser.add_argument("--plot", action='store_true', help="Plot transform (needs installed matplotlib package)")

args = parser.parse_args()
if not os.path.exists(args.input):
    parser.error("Input file '%s' not found"%args.input)

fs = args.sr

# build transform
scales = {'log':LogScale, 'lin':LinScale, 'mel':MelScale, 'oct':OctScale}
try:
    scale = scales[args.scale]
except KeyError:
    parser.error('Scale unknown (--scale option)')

scl = scale(args.fmin, args.fmax, args.bins, beyond=int(args.reducedform == 2))

slicq = NSGT_sliced(scl, args.sllen, args.trlen, fs, 
                    real=args.real, recwnd=args.recwnd, 
                    matrixform=args.matrixform, reducedform=args.reducedform, 
                    multithreading=args.multithreading,
                    multichannel=True
                    )

# Read audio data
sf = SndReader(args.input, sr=fs, chns=2)
signal = sf()

# duration of signal in s
dur = sf.frames/float(fs)

# total number of coefficients to represent input signal
ncoefs = int(sf.frames*slicq.coef_factor)

# read slices from audio file and mix down signal, if necessary at all
if not args.downmix_after:
    signal = ((np.mean(s, axis=0),) for s in signal)

# generator for forward transformation
c = slicq.forward(signal)

# add up overlapping coefficient slices
coefs = assemble_coeffs(c, ncoefs)

del sf # not needed any more

# compute magnitude spectrum
mindb = -100.
mls = np.abs(coefs)
# mix down multichannel
mls = np.mean(mls, axis=-1)
np.maximum(mls, 10**(mindb/20.), out=mls)

if args.mag_scale == 'dB':
    np.log10(mls, out=mls)
    mls *= 20.
elif args.mag_scale == 'log':
    np.log(mls, out=mls)
else:
    raise NotImplementedError("Magnitude scale '%s' not implemented."%args.magscale)

fs_coef = fs*slicq.coef_factor # frame rate of coefficients
mls_dur = len(mls)/fs_coef # final duration of MLS

if args.fps:
    # pool features in time
    coefs_per_sec = fs*slicq.coef_factor
    poolf = int(coefs_per_sec/args.fps+0.5) # pooling factor
    pooled_len = mls.shape[0]//poolf
    mls_dur *= (pooled_len*poolf)/float(len(mls))
    mls = mls[:pooled_len*poolf]
    poolfun = np.__dict__[args.fps_pooling]
    mls = poolfun(mls.reshape((pooled_len,poolf,)+mls.shape[1:]), axis=1)

times = np.linspace(0, mls_dur, endpoint=True, num=len(mls)+1)

# save file
if args.output:
    if args.reducedform == 2:
        frqs = slicq.frqs[1:-1]
        qs = slicq.q[1:-1]
    else:
        frqs = slicq.frqs
        qs = slicq.q
    
    data = {args.data_coefs: mls, args.data_times: times, args.data_frqs: frqs, args.data_qs: qs}
    if args.output.endswith('.pkl') or args.output.endswith('.pck'):
        import pickle
        with file(args.output, 'wb') as f:
            pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
    elif args.output.endswith('.npz'):
        np.savez(args.output, **data)
    elif args.output.endswith('.hdf5') or args.output.endswith('.h5'):
        import h5py
        with h5py.File(args.output, 'w') as f:
            for k,v in data.items():
                f[k] = v
    else:
        warn("Output file format not supported, skipping output.")

if args.plot:
    print("Plotting t*f space")
    import matplotlib.pyplot as pl
    mls_max = np.percentile(mls, 99.9)
    pl.imshow(mls.T, aspect=mls_dur/mls.shape[1]*0.2, interpolation='nearest', origin='bottom', vmin=mls_max-60., vmax=mls_max, extent=(0,mls_dur,0,mls.shape[1]))
    pl.show()



#! /usr/bin/env python
# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2021
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)

covered by the Artistic License 2.0
http://www.perlfoundation.org/artistic_license_2_0

--

Installation:

In the console (terminal application) change to the folder containing this readme.txt file.

To build the package run the following command:
python setup.py build

To install the package (with administrator rights):
sudo python setup.py install

--

Attention: some Cython versions also need the Pyrex module installed!

"""

import warnings
import os

try:
    import setuptools
except ImportError:
    warnings.warn("setuptools not found, resorting to distutils: unit test suite can not be run from setup.py")
    setuptools = None

setup_options = {}

if setuptools is None:
    from distutils.core import setup
    from distutils.extension import Extension
else:
    from setuptools import setup
    from setuptools.extension import Extension
    setup_options['test_suite'] = 'tests'
    
try:
    from Cython.Distutils import build_ext
except ImportError:
    build_ext = None
    
import numpy


if build_ext is None:
    cmdclass = {}
    ext_modules = []
else:
    cmdclass = {'build_ext': build_ext}
    ext_modules = [
                   Extension("nsgt._nsgtf_loop", ["nsgt/nsgtf_loop.pyx"]),
                   Extension("nsgt._nsigtf_loop", ["nsgt/nsigtf_loop.pyx"])
    ]

setup(
    name = "nsgt",
    version = "0.19",
    author = "Thomas Grill",
    author_email = "gr@grrrr.org",
    maintainer = "Thomas Grill",
    maintainer_email = "gr@grrrr.org",
    description = "Python implementation of Non-Stationary Gabor Transform (NSGT)",
    license = "Artistic License",
    keywords = "fourier gabor",
    url = "http://grrrr.org/nsgt",
    requires = ["numpy"],
    include_dirs = [numpy.get_include()],
    packages = ['nsgt'],
    cmdclass = cmdclass,
    ext_modules = ext_modules,
    classifiers = [
        "Development Status :: 4 - Beta",
        "Topic :: Scientific/Engineering :: Mathematics",
        "License :: OSI Approved :: Artistic License",
        "Programming Language :: Python"
    ],
    **setup_options
)



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)

--
Original matlab code copyright follows:

AUTHOR(s) : Monika Dörfler, Gino Angelo Velasco, Nicki Holighaus, 2010-2011

COPYRIGHT : (c) NUHAG, Dept.Math., University of Vienna, AUSTRIA
http://nuhag.eu/
Permission is granted to modify and re-distribute this
code in any manner as long as this notice is preserved.
All standard disclaimers apply.

"""

from .nsgfwin_sl import nsgfwin
from .nsdual import nsdual
from .nsgtf import nsgtf
from .nsigtf import nsigtf
from .util import calcwinrange
from .fscale import OctScale
from math import ceil

class NSGT:
    def __init__(self, scale, fs, Ls, real=True, matrixform=False, reducedform=0, multichannel=False, measurefft=False, multithreading=False, dtype=float):
        assert fs > 0
        assert Ls > 0
        assert 0 <= reducedform <= 2
        
        self.scale = scale
        self.fs = fs
        self.Ls = Ls
        self.real = real
        self.measurefft = measurefft
        self.multithreading = multithreading
        self.reducedform = reducedform
        
        self.frqs,self.q = scale()

        # calculate transform parameters
        self.g,rfbas,self.M = nsgfwin(self.frqs, self.q, self.fs, self.Ls, sliced=False, dtype=dtype)

        if real:
            assert 0 <= reducedform <= 2
            sl = slice(reducedform,len(self.g)//2+1-reducedform)
        else:
            sl = slice(0,None)

        # coefficients per slice
        self.ncoefs = max(int(ceil(float(len(gii))/mii))*mii for mii,gii in zip(self.M[sl],self.g[sl]))        

        if matrixform:
            if self.reducedform:
                rm = self.M[self.reducedform:len(self.M)//2+1-self.reducedform]
                self.M[:] = rm.max()
            else:
                self.M[:] = self.M.max()
    
        if multichannel:
            self.channelize = lambda s: s
            self.unchannelize = lambda s: s
        else:
            self.channelize = lambda s: (s,)
            self.unchannelize = lambda s: s[0]

        # calculate shifts
        self.wins,self.nn = calcwinrange(self.g, rfbas, self.Ls)
        # calculate dual windows
        self.gd = nsdual(self.g, self.wins, self.nn, self.M)
        
        self.fwd = lambda s: nsgtf(s, self.g, self.wins, self.nn, self.M, real=self.real, reducedform=self.reducedform, measurefft=self.measurefft, multithreading=self.multithreading)
        self.bwd = lambda c: nsigtf(c, self.gd, self.wins, self.nn, self.Ls, real=self.real, reducedform=self.reducedform, measurefft=self.measurefft, multithreading=self.multithreading)
        
    @property
    def coef_factor(self):
        return float(self.ncoefs)/self.sl_len
    
    @property
    def slice_coefs(self):
        return self.ncoefs
    

    def forward(self, s):
        'transform'
        s = self.channelize(s)
        c = list(map(self.fwd, s))
        return self.unchannelize(c)

    def backward(self, c):
        'inverse transform'
        c = self.channelize(c)
        s = list(map(self.bwd,c))
        return self.unchannelize(s)
    
class CQ_NSGT(NSGT):
    def __init__(self, fmin, fmax, bins, fs, Ls, real=True, matrixform=False, reducedform=0, multichannel=False, measurefft=False, multithreading=False):
        assert fmin > 0
        assert fmax > fmin
        assert bins > 0
        
        self.fmin = fmin
        self.fmax = fmax
        self.bins = bins

        scale = OctScale(fmin, fmax, bins)
        NSGT.__init__(self, scale, fs, Ls, real, matrixform=matrixform, reducedform=reducedform, multichannel=multichannel, measurefft=measurefft, multithreading=multithreading)



import numpy as np
from nsgt import NSGT, OctScale
import unittest

class TestNSGT(unittest.TestCase):

    def test_oct(self):
        siglen = int(10**np.random.uniform(4,6))
        sig = np.random.random(siglen)
        fmin = np.random.random()*200+20
        fmax = np.random.random()*(22048-fmin)+fmin
        obins = np.random.randint(24)+1
        scale = OctScale(fmin,fmax,obins)
        nsgt = NSGT(scale, fs=44100, Ls=len(sig))
        c = nsgt.forward(sig)
        s_r = nsgt.backward(c)
        self.assertTrue(np.allclose(sig, s_r, atol=1e-07))

def load_tests(*_):
    # seed random generators for unit testing
    np.random.seed(666)

    test_cases = unittest.TestSuite()
    for _ in range(100):
        test_cases.addTest(TestNSGT('test_oct'))
    return test_cases

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2015
http://grrrr.org/nsgt

Austrian Research Institute for Artificial Intelligence (OFAI)
AudioMiner project, supported by Vienna Science and Technology Fund (WWTF)
"""

import numpy as np
from .util import hannwin
from .reblock import reblock
from itertools import chain, cycle

def makewnd(sl_len, tr_area):
    hhop = sl_len//4
    htr = tr_area//2
    # build window function within one slice (centered with transition areas around sl_len/4 and 3*sl_len/4    
    w = hannwin(2*tr_area)  # window is shifted
    tw = np.empty(sl_len, dtype=float)
    tw[:hhop-htr] = 0
    tw[hhop-htr:hhop+htr] = w[tr_area:]
    tw[hhop+htr:3*hhop-htr] = 1
    tw[3*hhop-htr:3*hhop+htr] = w[:tr_area]
    tw[3*hhop+htr:] = 0
    return tw

def slicing(f, sl_len, tr_area):
    if tr_area%2 != 0:
        raise ValueError("Transition area 'tr_area' must be modulo 2")
    if sl_len%4 != 0:
        raise ValueError("Slice length 'sl_len' must be modulo 4")
    
    hhop = sl_len//4  # half hopsize

    tw = makewnd(sl_len, tr_area)
    # four parts of slice with centered window function
    tw = [tw[o:o+hhop] for o in range(0, sl_len, hhop)]
    
    # stream of hopsize/2 blocks with leading and trailing zero blocks
    fseq = reblock(f, hhop, dtype=float, fulllast=True, padding=0., multichannel=True)
    
    # get first block to deduce number of channels
    fseq0 = next(fseq)
    chns = len(fseq0)
    pad = np.zeros((chns,hhop), dtype=fseq0.dtype)
    # assemble a stream of front padding, already retrieved first block, the block stream and some tail padding
    fseq = chain((pad,pad,fseq0), fseq, (pad,pad,pad))

    slices = [[slice(hhop*((i+3-k*2)%4), hhop*((i+3-k*2)%4+1)) for i in range(4)] for k in range(2)]
    slices = cycle(slices)
    
    past = []
    for fi in fseq:
        past.append(fi)
        if len(past) == 4:
            f_slice = np.empty((chns,sl_len), dtype=fi.dtype)
            sl = next(slices)
            for sli,pi,twi in zip(sl, past, tw):
                f_slice[:,sli] = pi    # signal
                f_slice[:,sli] *= twi  # multiply with part of window function
            yield f_slice
            past = past[2:]  # pop the two oldest slices



#! /usr/bin/env python 
# -*- coding: utf-8

"""
Python implementation of Non-Stationary Gabor Transform (NSGT)
derived from MATLAB code by NUHAG, University of Vienna, Austria

Thomas Grill, 2011-2020
http://grrrr.org/nsgt
"""

import numpy as np
from scikits.audiolab import Sndfile, Format
import os

from nsgt import NSGT_sliced, LogScale, LinScale, MelScale, OctScale
from nsgt.reblock import reblock

def cputime():
    utime, stime, cutime, cstime, elapsed_time = os.times()
    return utime


from argparse import ArgumentParser
parser = ArgumentParser()

parser.add_argument("input", type=str, help="input file")
parser.add_argument('-o', "--output", type=str, help="output audio file")
parser.add_argument('-L', "--length", type=int, default=0, help="maximum length of signal (default=%(default)s)")
parser.add_argument('-f', "--fmin", type=float, default=50, help="minimum frequency (default=%(default)s)")
parser.add_argument('-F', "--fmax", type=float, default=22050, help="maximum frequency (default=%(default)s)")
parser.add_argument('-s', "--scale", choices=('oct','log','mel'), default='log', help="frequency scale (oct,log,lin,mel)")
parser.add_argument('-b', "--bins", type=int, default=50, help="frequency bins (total or per octave, default=%(default)s)")
parser.add_argument("--sllen", type=int, default=2**16, help="slice length (default=%(default)s)")
parser.add_argument("--trlen", type=int, default=4096, help="transition area (default=%(default)s)")
parser.add_argument('-r', "--real", action='store_true', help="assume real signal")
parser.add_argument('-m', "--matrixform", action='store_true', help="use regular time division (matrix form)")
parser.add_argument('-l', "--reducedform", action='count', default=0, help="if real==1: omit bins for f=0 and f=fs/2 (lossy=1), or also the transition bands (lossy=2)")
parser.add_argument('-w', "--recwnd", action='store_true', help="use reconstruction window")
parser.add_argument('-t', "--multithreading", action='store_true', help="use multithreading")
parser.add_argument('-p', "--plot", action='store_true', help="plot transform (needs installed matplotlib and scipy packages)")

args = parser.parse_args()
if not os.path.exists(args.input):
    parser.error("Input file '%s' not found"%args.input)

# Read audio data
sf = Sndfile(args.input)
fs = sf.samplerate
s = sf.read_frames(sf.nframes)
if sf.channels > 1: 
    s = np.mean(s, axis=1)
    
if args.length:
    s = s[:args.length]

scales = {'log':LogScale,'lin':LinScale,'mel':MelScale,'oct':OctScale}
try:
    scale = scales[args.scale]
except KeyError:
    parser.error('scale unknown')

scl = scale(args.fmin, args.fmax, args.bins)
slicq = NSGT_sliced(scl, args.sllen, args.trlen, fs, 
                    real=args.real, recwnd=args.recwnd, 
                    matrixform=args.matrixform, reducedform=args.reducedform, 
                    multithreading=args.multithreading
                    )

t1 = cputime()

signal = (s,)

# generator for forward transformation
c = slicq.forward(signal)

# realize transform from generator
c = list(c)

# generator for backward transformation
outseq = slicq.backward(c)

# make single output array from iterator
s_r = next(reblock(outseq, len(s), fulllast=False))
s_r = s_r.real

t2 = cputime()

norm = lambda x: np.sqrt(np.sum(np.abs(np.square(np.abs(x)))))
rec_err = norm(s-s_r)/norm(s)
print("Reconstruction error: %.3e"%rec_err)
print("Calculation time: %.3fs"%(t2-t1))

# Compare the sliced coefficients with non-sliced ones
if False:
    # not implemented yet!
    test_coeff_quality(c, s, g, shift, M, options.sl_len, len(s))

if args.output:
    print("Writing audio file '%s'"%args.output)
    sf = Sndfile(args.output, mode='w', format=Format('wav','pcm24'), channels=1, samplerate=fs)
    sf.write_frames(s_r)
    sf.close()
    print("Done")

if args.plot:
    print("Plotting t*f space")
    import matplotlib.pyplot as pl
    tr = np.array([[np.mean(np.abs(cj)) for cj in ci] for ci in c])
    pl.imshow(np.log(np.flipud(tr.T)+1.e-10), aspect=float(tr.shape[0])/tr.shape[1]*0.5, interpolation='nearest')
    pl.show()
